import java.io.*;
import java.util.*;
import java.net.*;

public class ClientHandler {

    private Compiler compiler;
    private CompileServer server;
    private Statistics statistics;

    public ClientHandler(CompileServer server, Compiler compiler, Statistics statistics) {
        this.compiler = compiler;
        this.server = server;
        this.statistics = statistics;
    }

    public static class InteractionException extends Exception {
        public InteractionException(String message) {
            super(message);
        }
    }

    public void handle(Socket socket) {
        BufferedReader in = null;
        PrintStream out = null;
        try {
            in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            out = new PrintStream(socket.getOutputStream());
            this.statistics.addConnect();
            out.println("JavaCompileServer: HELLO: Enter passcode.");
            String input = in.readLine();
            if (input == null) {
                out.println("JavaCompileServer: ERROR: invalid passcode");
                throw new InteractionException("No passcode provided.");
            }
            if (!this.server.isValidPassphrase(input)) {
                out.println("JavaCompileServer: ERROR: invalid passcode");
                throw new InteractionException("Incorrect passcode: '" + input + "'");
            }
            this.statistics.addAuthorizedConnect();

            String command = in.readLine();
            if (command == null) throw new InteractionException("no request made.");
            command = command.toLowerCase();
            this.statistics.addRequest();
            if (command.equals("compile")) {
                this.statistics.addCompileRequest();
                handleCompile(in, out);
            } else if (command.equals("statistics")) {
                this.statistics.addStatisticsRequest();
                this.server.printInfo(out);
                this.statistics.print(out);
            } else if (command.equals("signature")) {
                this.server.printSignature(out);
            } else {
                out.println("JavaCompileServer: ERROR Unknown command '" + command + "'");
                throw new InteractionException("invalid request + '" + command + "'");
            }
        } catch (IOException ex) {
            this.server.log("ERROR: IO error while dealing with client: " + ex);
        } catch (ClientHandler.InteractionException ex) {
            this.server.log("ERROR: Client doing something stupid: " + ex);
        } catch (Compiler.CompilationException ex) {
            this.server.log("ERROR: Error when trying to compile: " + ex);
        } finally {
            try {
                if (out != null) out.close();
                if (in != null) in.close();
            } catch (IOException ex) {
                // ignore
            }
        }
    }

    private void handleCompile(BufferedReader in, PrintStream out) throws IOException, InteractionException, Compiler.CompilationException {
        long startTime = System.currentTimeMillis();
        // user.dir
        String userDir = in.readLine();
        if (userDir == null) throw new InteractionException("HANDLER: Couldn't read 'current-directory' string.");
        File userDirFile = new File(userDir);
        if (!userDirFile.isAbsolute()) {
            out.println("JavaCompileServer: ERROR: Value for 'current-directory' must be an absolute path.");
            throw new InteractionException("HANDLER: Given 'current-directory' isn't absolute.");
        }
        boolean outputDirectorySpecified = false;
        boolean classpathSpecified = false;
        boolean sameDirectory = userDirFile.equals(this.server.getCurrentDirectory());
        if (sameDirectory) {
            outputDirectorySpecified = true;
            classpathSpecified = true;
        }
        LinkedList arguments = new LinkedList();
        boolean ignoreNext = false;
        while (true) {
            String arg = in.readLine();
            if (arg == null) throw new InteractionException("EOF in the middle of the argument list.");
            if (arg.equals("")) break;
            if (!ignoreNext) {
                if (!arg.startsWith("-") && !sameDirectory) {
                    // it's a directory/file of some sort, right?
                    if (!sameDirectory) arg = fixArgument(arg, userDirFile);
                } else {
                    if (arg.equals("-d")) {
                        outputDirectorySpecified = true;
                    } else if (arg.equals("-classpath") || arg.equals("-cp")) {
                        classpathSpecified = true;
                    } else if (arg.equals("-encoding")) {
                        ignoreNext = true;
                    } else if (arg.equals("-target")) {
                        ignoreNext = true;
                    } else if (arg.equals("-source")) {
                        ignoreNext = true;
                    }
                }
                arguments.add(arg);
            } else {
                ignoreNext = false;
            }
        }
        if (arguments.size() > 0) {
            if (!classpathSpecified) {
                arguments.addFirst(userDirFile.getAbsolutePath());
                arguments.addFirst("-classpath");
            }
            if (!outputDirectorySpecified) {
                arguments.addFirst(userDirFile.getAbsolutePath());
                arguments.addFirst("-d");
            }
        }
        String[] argArray = (String[]) arguments.toArray(new String[arguments.size()]);
        for (int i = 0; i < argArray.length; i++) {
            out.println("JavaCompileServer: DEBUG: arg[" + i + "] = '" + argArray[i] + "'");
        }
        out.println("JavaCompileServer: OK: Compiled results follow.");
        int result;
        try {
            result = compiler.compile(out, argArray);
        } catch (Compiler.CompilationException ex) {
            out.println("JavaCompileServer: ERROR: (server error) exception when compiling.");
            out.println("JavaCompileServer: DEBUG: (server error) exception: " + ex);
            server.log("ERROR: exception while compiling: " + ex);
            throw ex;
        }
        out.println("JavaCompileServer: OK: return value = " + result);
        out.flush();
        long elapsed = System.currentTimeMillis() - startTime;
        this.statistics.addCompilationTime(elapsed, result);
    }

    private String fixArgument(String arg, File userDirFile) {
        int pos = 0;
        int length = arg.length();
        StringBuffer newArg = null;
        while (pos < length) {
            int div = arg.indexOf(File.pathSeparatorChar, pos);
            if (div < 0) div = length;
            if ((div - pos) > 0) {
                String part = arg.substring(pos, div);
                File file = new File(part);
                if (!file.isAbsolute()) {
                    file = new File(userDirFile, part);
                }
                part = file.getAbsolutePath();
                if (newArg == null) {
                    newArg = new StringBuffer();
                    newArg.append(part);
                } else {
                    newArg.append(File.pathSeparatorChar);
                    newArg.append(part);
                }
            } 
            pos = div + 1;
        }
        return (newArg.toString());
    }

}
