import java.net.*;
import java.util.*;
import java.text.*;
import java.io.*;
import java.lang.reflect.Method;

public class CompileServer implements Runnable {

    private PrintStream out, err, log;

    private ServerSocket listener;

    private String passphrase;
    public static final int PASSCODE_MIN_LENGTH = 1;

    private File portFile, killFile;
    private FileDeleter portFileDeleter;
    private Options options;
    private File currentDirectory;

    private Compiler compiler;

    Statistics statistics;

    private static StartupCanceller cancel = new StartupCanceller();

    public CompileServer(Options options, PrintStream out, PrintStream err) throws StartupCanceller {
        if (out == null) throw new NullPointerException("Argument 'out' to CompileServer constructor cannot be null.");
        this.out = out;
        if (err == null) throw new NullPointerException("Argument 'err' to CompileServer constructor cannot be null.");
        this.err = err;
        if (options == null) throw new NullPointerException("Argument 'options' to CompileServer constructor cannot be null.");
        this.options = options;

        this.currentDirectory = new File(System.getProperty("user.dir"));

        this.statistics = new Statistics();

        this.portFile = new File(this.options.portFile);
        try {
            if (!this.portFile.createNewFile()) {
                out("Warning: Unable to create port file.  Checking for another server.");
                if (isServerRunning(this.portFile)) {
                    err("ERROR: Port file already exists and another server seems to be");
                    err("ERROR: running at the specified port.");
                    err("ERROR: Stop the other server and try again.  If you are sure");
                    err("ERROR: that no other server is running, delete the port-file");
                    err("ERROR: at \"" + this.portFile.getPath() + "\" and try again.");
                    throw cancel;
                }
            }
        } catch (IOException ex) {
            err("ERROR: Unable to create port file \"" + this.portFile.getPath() + "\": " + ex);
            throw cancel;
        }

        try {
            out("Loading compiler...");
            try {
                this.compiler = Compiler.Loader.findCompiler();
                out("  * OK [version='" + this.compiler.toString() + "']");
            } catch (Exception ex) {
                err("ERROR: couldn't load compiler: " + ex);
                err("ERROR: Are you sure you properly configured the classpath?");
                throw cancel;
            }

            out("Creating server socket...");
            try {
                this.listener = new ServerSocket(this.options.port);
                this.options.port = listener.getLocalPort();
                try {
                    this.listener.setSoTimeout(1000);
                } catch (SocketException ex) {
                    err("    NOTICE: unable to set socket timeout value: " + ex);
                }
            } catch (Exception ex) {
                err("ERROR: unable to create server socket: " + ex);
                throw cancel;
            }
            out("  * OK [port=" + this.options.port + "]");

            out("Writing to port file [" + this.options.portFile + "]...");
            try {
                File parent = this.portFile.getAbsoluteFile().getParentFile();
                if (parent != null && !parent.exists()) parent.mkdirs();
                PrintStream portFileStream = new PrintStream(new FileOutputStream(this.portFile));
                portFileStream.println(this.options.port);
                portFileStream.close();
                this.portFileDeleter = new FileDeleter(this.portFile);
                registerShutdownHook(this.portFileDeleter);
            } catch (Exception ex) {
                err("ERROR: unable to write to port file: " + ex);
                throw (cancel);
            }
            out("  * OK");



            out("Checking for kill file [" + this.options.killFile + "]...");
            this.killFile = new File(this.options.killFile);
            if (this.killFile.exists()) {
                out("    NOTICE: kill file already exists.  This could have been");
                out("            caused by an unclean server shut down.");
                out("    > Deleting old kill-file...");
                if (!this.killFile.delete()) {
                    if (this.killFile.exists()) { // check again.  why?  for fun!
                        err("ERROR: Unable to delete existing kill-file.");
                        err("ERROR: Delete it yourself and try again.");
                        throw cancel;
                    }
                }
            }
            out("  * OK");

            out("Reading from passphrase file [" + this.options.passFile + "]...");
            try {
                BufferedReader passFileStream = new BufferedReader(new FileReader(this.options.passFile));
                this.passphrase = passFileStream.readLine();
                if (this.passphrase == null || this.passphrase.length() < PASSCODE_MIN_LENGTH) {
                    err("ERROR: passphrase file's first line must have at least " + PASSCODE_MIN_LENGTH + " characters.");
                    throw cancel;
                }
                passFileStream.close();
            } catch (Exception ex) {
                err("ERROR: unable to read passphrase from pass-file: " + ex);
                throw cancel;
            }
            out("  * OK [" + this.passphrase.length() + " chars]");

            
            out("Opening log file [" + this.options.logFile + "]...");
            PrintStream log;
            try {
                this.completeFormat = new SimpleDateFormat("EEE, dd MMM yyyy, HH:mm:ss, z");
                this.logTimeFormat = new SimpleDateFormat("[HH:mm:ss] ");
                this.logDateFormat = new SimpleDateFormat("EEE, dd MMM yyyy");

                File logFile = new File(this.options.logFile);
                File parent = logFile.getAbsoluteFile().getParentFile();
                if (parent != null && !parent.exists()) parent.mkdirs();
                log = new PrintStream(new FileOutputStream(logFile));
                this.logDate = new Date();
                this.startDate = new Date();
                this.log = log;
                this.log.println(completeFormat.format(logDate));
                this.log.println("JavaCompileServer started on port " + this.options.port + ".");
            } catch (IOException ex) {
                err("ERROR: unable to create log file: " + ex);
                throw cancel;
            }
            out("  * OK");

            // write signal
            try {
                File signalFile = new File(this.options.signalFile);
                File parent = signalFile.getAbsoluteFile().getParentFile();
                if (parent != null && !parent.exists()) parent.mkdirs();
                if (!signalFile.createNewFile()) throw new Exception();
            } catch (Exception ex) {
                err("ERROR: unable to create signal file.");
                throw (cancel);
            }
        } catch (StartupCanceller sc) {
            try {
                this.portFileDeleter.run();
            } catch (Exception ex) {
                // ignore
            }
            throw sc;  // rethrow it
        }

    }

    public boolean isValidPassphrase(String string) {
        return (this.passphrase.equals(string));
    }

    public File getCurrentDirectory() {
        return (this.currentDirectory);
    }

    private boolean isServerRunning(File portFile) {
        int port;
        try {
            BufferedReader in = new BufferedReader(new FileReader(portFile));
            String portString = in.readLine();
            port = Integer.parseInt(portString);
            if (port < 1 || port > 65535) throw new NumberFormatException();
        } catch (FileNotFoundException ex) {
            out("      NOTICE: That's weird.  An old port file existed just moments");
            out("              ago and now now it's gone.  Oh well.");
            return (false);
        } catch (NumberFormatException ex) {
            out("      NOTICE: Existing port file doesn't seem to contain a valid");
            out("              port number.  Can't check if another server is");
            out("              already running.");
            return (false);
        } catch (Exception ex) {
            out("      NOTICE: Couldn't read from existing port file.  Can't check");
            out("              if another server is already running.");
            return (false);
        }

        boolean running = false;
        try {
            Socket socket = new Socket(InetAddress.getLocalHost(), port);
            BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            try {
                socket.setSoTimeout(5000);
            } catch (SocketException ex) {
                out("     NOTICE: Unable to limit wait time on socket read.  The program");
                out("             might get stuck here.");
            }
            try {
                String input = in.readLine();
                running = input.startsWith("JavaCompileServer");
            } catch (InterruptedIOException ex) {
                running = false;
            }
            if (!running) {
                out("     > No JavaCompileServer found.  Proceeding to overwrite old");
                out("       port-file.");
            }
            socket.close();
            return (running);
        } catch (IOException ex) {
            out("     > No server found.  Proceeding to overwrite old port-file.");
            return (false);
        }
    }

    public void run() {
        Statistics statistics = new Statistics();
        ClientHandler handler = new ClientHandler(this, this.compiler, statistics);

        outer: while (true) {
            Socket socket = null;
            try {
                while (socket == null) {
                    try {
                        socket = this.listener.accept();
                    } catch (InterruptedIOException ex) { 
                        // check for killFile
                        if (this.killFile.exists()) {
                            // time to shutdown
                            this.killFile.delete();
                            break outer;
                        }
                    }
                }
            } catch (IOException ex) {
                log("ERROR: Error while accepting connection: " + ex);
            }

            handler.handle(socket);
        }

        try {
            this.listener.close();
        } catch (Exception ex) {
            // ignore
        }

        this.portFileDeleter.run();
    }

    private void registerShutdownHook(FileDeleter deleter) {
        try {
            Method registerationMethod = Runtime.class.getMethod("addShutdownHook", new Class[] {Thread.class});
            Runtime runtime = Runtime.getRuntime();
            registerationMethod.invoke(runtime, new Object[]{new Thread(deleter)});
        } catch (Exception ex) {
            // ignore
        }
    }

    private static class FileDeleter implements Runnable {
        private File file;
        private long lastModified;
        public FileDeleter(File file) {
            this.file = file;
            this.lastModified = file.lastModified();
        }
        public synchronized void run() {
            if (this.file.exists() && this.file.lastModified() == this.lastModified) {
                this.file.delete();
            }
        }
    }

/*
    public void outp(String message) {
        this.out.print(message); this.out.flush();
    }
    */

    private DateFormat completeFormat;
    private DateFormat logTimeFormat;
    private DateFormat logDateFormat;
    private String previousDate;
    private Date logDate;
    private Date startDate;

    public void log(String message) {
        this.logDate.setTime(System.currentTimeMillis());
        String currentDate = this.logDateFormat.format(this.logDate);
        if (previousDate == null || !previousDate.equals(currentDate)) {
            this.log.print(">> ");
            this.log.println(currentDate);
            this.previousDate = currentDate;
        }
        this.log.print(this.logTimeFormat.format(this.logDate));
        this.log.println(message); this.log.flush();
    }

    public void out(String message) {
        this.out.println(message); this.out.flush();
    }

    public void err(String message) {
        this.err.println(message); this.err.flush();
    }

    public int getPort() {
        return (this.options.port);
    }

    public void printInfo(PrintStream out) {
        out.println("Program [name/vendor/version]: JavaCompileServer/cakoose/0.0.1");
        out.println("Started: " + this.completeFormat.format(this.startDate));
    }

    public void printSignature(PrintStream out) {
        out.println("JavaCompileServer/cakoose/0.0.1");
    }

    public static void main(String[] args) {
        Options options = null;
        try {
            options = new Options(args);
        } catch (Options.OptionsException ex) {
            System.err.println(ex);
            System.err.println(Options.USAGE);
            System.exit(1);
        }
        try {
            CompileServer server = new CompileServer(options, System.out, System.err);
            server.run();
        } catch (StartupCanceller ex) {
            // don't print anything, this is thrown only when the error message is already printed
            System.exit(1);
        }
    }

    public static class StartupCanceller extends Throwable { }

}


