import java.lang.reflect.*;
import java.io.*;

public interface Compiler {

    public static class CompilationException extends Exception {
        private Exception cause;
        public CompilationException(Exception cause) {
            this.cause = cause;
        }
        public String toString() {
            return ("CompilationException: " + this.cause.toString());
        }
        public void printStackTrace(PrintStream out) {
            this.cause.printStackTrace(out);
        }
    }

    public int compile(OutputStream out, String[] arguments) throws CompilationException;

    public static class Loader {
        public static Compiler findCompiler() {
            try {
                return (new Version_5());
            } catch (Exception ex) {}

            try {
                return (new Version_1_4());
            } catch (Exception ex) {}

            try {
                return (new Version_1_3());
            } catch (Exception ex) {}

            try {
                return (new Version_1_2());
            } catch (Exception ex) {}

            return (null);
        }

    }

    /**
     * Generic class that'll look up the appropriate method for you.  Works
     * with both static and instance methods.
     */
    public static abstract class SimpleReflector implements Compiler {
        protected Object instance;
        protected Method method;
        protected String toString;
        public SimpleReflector(String toString) {
            this.toString = toString;
        }
        public String toString() {
            return (this.toString);
        }
        protected void initialize(String className, Class[] constructorParamTypes,
                             Object[] constructorParams, String methodName,
                             Class[] methodParamTypes) throws Exception {
            Class c = Class.forName(className);
            this.method = c.getMethod(methodName, methodParamTypes);

            int modifiers = this.method.getModifiers();
            if ((modifiers & Modifier.PUBLIC) == 0)
                throw new IllegalArgumentException("Compilation method is not public.");
            if ((modifiers & Modifier.ABSTRACT) != 0)
                throw new IllegalArgumentException("Compilation method is abstract.");

            if (constructorParamTypes != null) {
                Constructor constructor = c.getConstructor(constructorParamTypes);
                this.instance = constructor.newInstance(constructorParams);
            } else {
                if ((modifiers & Modifier.STATIC) == 0) {
                    throw new IllegalArgumentException("No constructor info provided, but method is not static.");
                }
            }

            Class returnType = this.method.getReturnType();
            if (!returnType.isInstance(Boolean.class) && !returnType.isInstance(Integer.class)
                        && !returnType.equals(boolean.class) && !returnType.equals(int.class))
                throw new IllegalArgumentException("Unnacceptable return type: " + returnType.toString());
        }
        protected int compile(Object[] methodParams) throws CompilationException {
            try {
                Object returnValue = this.method.invoke(instance, methodParams);
                if (returnValue instanceof Boolean) {
                    return (((Boolean)returnValue).booleanValue() ? 0 : 1);
                } else {
                    return (((Integer)returnValue).intValue());
                }
            } catch (Exception ex) {
                throw new CompilationException(ex);
            }
        }
    }

    /**
     * Generic class that'll redirect stdout and stderr before calling the
     * compiler.
     */
    public abstract class StreamSwitcher extends SimpleReflector {
        private ReusableOutputStream redirect;
        private PrintStream output;
        protected StreamSwitcher(String name) {
            super(name);
            this.redirect = new ReusableOutputStream();
            this.output = new PrintStream(this.redirect);
        }
        public int compile(OutputStream pout, String[] arguments) throws CompilationException {
            PrintStream oldOut = System.out;
            PrintStream oldErr = System.err;
            try {
                this.redirect.setOutputStream(pout);
                System.setOut(this.output);
                System.setErr(this.output);
                Object[] methodArguments = { arguments };
                return super.compile(methodArguments);
            } catch (CompilationException ex) {
                throw ex;
            } catch (Exception ex) {
                throw new CompilationException(ex);
            } finally {
                System.setOut(oldOut);
                System.setErr(oldErr);
            }
        }
    }

    // JDK 1.2
    public static class Version_1_2 extends SimpleReflector {
        private static final String className = "sun.tools.javac.Main";
        private static final String methodName = "compile";
        private static final Class[] constructorParamTypes = { OutputStream.class, String.class };
        private static final Class[] methodParamTypes = { String[].class };
        private ReusableOutputStream redirect;
        public Version_1_2() throws Exception {
            super("JDK 1.2");
            this.redirect = new ReusableOutputStream();
            Object[] constructorParams = { this.redirect, "javac" };
            super.initialize(className, constructorParamTypes, constructorParams, methodName, methodParamTypes);
        }
        public int compile(OutputStream out, String[] compilerArguments) throws CompilationException {
            this.redirect.setOutputStream(out);
            Object[] methodParams = { compilerArguments };
            return (super.compile(methodParams));
        }

    }

    // JDK 1.3
    public static class Version_1_3 extends StreamSwitcher {
        private static final String className = "com.sun.tools.javac.Main";
        private static final String methodName = "compile";
        private static final Class[] constructorParamTypes = {};
        private static final Class[] constructorParams = {};
        private static final Class[] methodParamTypes = { String[].class };
        public Version_1_3() throws Exception {
            super("JDK 1.3");
            super.initialize(className, constructorParamTypes, constructorParams, methodName, methodParamTypes);
        }
    }

    // JDK 1.4
    public static class Version_1_4 extends StreamSwitcher {
        private static final String className = "com.sun.tools.javac.Main";
        private static final String methodName = "compile";
        private static final Class[] constructorParamTypes = {};
        private static final Class[] constructorParams = {};
        private static final Class[] methodParamTypes = { String[].class };
        public Version_1_4() throws Exception {
            super("JDK 1.4/5.0");
            super.initialize(className, null, null, methodName, methodParamTypes);
        }
    }

    // JDK 5.0
    public static class Version_5 extends SimpleReflector {
        private static final String className = "com.sun.tools.javac.Main";
        private static final String methodName = "compile";
        private static final Class[] constructorParamTypes = {};
        private static final Class[] constructorParams = {};
        private static final Class[] methodParamTypes = { String[].class, PrintWriter.class };
        public Version_5() throws Exception {
            super("JDK 5.0");
            super.initialize(className, null, null, methodName, methodParamTypes);
        }
        public int compile(OutputStream out, String[] compilerArguments) throws CompilationException {
            Object[] methodParams = { compilerArguments, new PrintWriter(out) };
            return (super.compile(methodParams));
        }
    }

}
